/**
LodePNG version 20070102

Copyright (c) 2005-2007 Lode Vandevenne
All rights reserved.

License:

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of Lode Vandevenne nor the names of his contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Authors:
Lode Vandevenne - original version in C++.<br>
Lutger Blijdestijn - port to D.

Port:
Port to D by Lutger Blijdestijn on 5 januari 2007.<br>
The port is close to the original, but the api has changed. I have taken out the class and used a procedural api instead.
Implementation-wise zlib routines are replaced by std.zlib.
Here follows part of the original documentation, slightly modified to reflect the D port.<br>
You can email me at: lutger dot blijdenstijn at gmail dot com

About:
PNG is a file format to store raster images losslessly with good compression, supporting different color types. It's patent-free.
Because the compression is lossless and it supports an alpha channel, PNG is a very good file format to store textures and tilesets
for computer games.

LodePNG is a PNG codec according to the Portable Network Graphics (PNG) Specification (Second Edition) - W3C Recommendation
10 November 2003. The specification can be found on line at http://www.w3.org/TR/2003/REC-PNG-20031110. The most recent version
of LodePNG can currently be found at http://student.kuleuven.be/~m0216922/pngloader/ However, this location is temporary and
won't exist anymore in 2007.

LodePNG exists out of the source code file lodepng.d

LodePNG is simple but only supports the basic requirements. In the tradeoff
between functionality and simplicity, LodePNG is on the side of simplicity.
To achieve this simplicity, the following design choices were made: There
are no dependencies on any external library. To decode PNGs, there's a Decoder
class (edit: in D port no class) that can convert any PNG file data into an RGBA image buffer with a single
function call.

This all makes LodePNG suitable for loading textures in games, raytracers,
intros, or for loading images into programs that require them only for simple
usage. It's less suitable for full fledged image editors, loading PNGs over
network (since this decoder requires all the image data to be available before
the decoding can begin), life-critical systems, ... Even though it contains
a conformant decoder and encoder, it's still not a conformant editor,
because unknown chunks are discarded.

Because LodePNG is BSD licensed, it's allowed to use these source files in any
project, as long as you include the above copyright message and conditions
somewhere in the release of your program.

Features:
The following features are supported by the decoder:
<UL>
  <li> conformant decoding of PNGs with any color type, bit depth and interlace mode</li>
  <li> converting the result to 32-bit RGBA pixel data</li>
  <li> decoding a png image from a buffer.</li>
  <li> support for translucent PNG's, including translucent palettes and color key</li>
  <li> zlib decompression (inflate) (via phobos in D)</li>
  <li> 64 bit color (untested due to lack of such images)</li>
  <li> the following chunks are interpreted by the decoder
  <ul>
    <li>IHDR (header information)</li>
    <li>PLTE (color palette)</li>
    <li>IDAT (pixel data)</li>
    <li>IEND (the final chunk)</li>
    <li>tRNS (transparency for palettized images)</li>
    <li>bKGD (suggested background color)</li>
    </ul>
   </li>
</UL>

<b>Features not supported:</b><br>
The following features are _not_ supported.
<ul>
<li> editing a PNG image (unless you use the decoder, then edit it, then use the
    encoder, but ignored chunks will then be gone from the original image)
<li> CRC checks of chunks
<li> ADLER32 checksum of zlib data
<li> partial loading. All data must be available and is processed in one call.
<li> The following optional chunk types are not interpreted by the decoder <ul>
    <li>cHRM (device independent color info)
    <li>gAMA (device independent color info)
    <li>iCCP (device independent color info)
    <li>sBIT (original number of significant bits)
    <li>sRGB (device independent color info)
    <li>tEXt (textual information)
    <li>zTXt (compressed textual information)
    <li>iTXt (international textual information)
    <li>hIST (palette histogram)
    <li>pHYs (physical pixel dimensions)
    <li>sPLT (suggested reduced palette)
    <li>tIME (last image modification time)
    </ul>
</ul>

Usage:
The basic steps of decoding a PNG file with LodePNG are as follows:
<OL>
<LI> Load the PNG image file from harddisk and store the bytes of the file in an ubyte[] with the same size as the file.
<LI> Declare another ubyte[] that will contain the pixel data.
<LI> Call the decode function of it with the above arrays as parameters and store the resulting Info struct.
<LI> Check for errors with in Info.error. (non-zero == error)
<LI> If no error occurred, the raw pixel data is now stored in the output array.
</OL>

To decode a PNG file, you can use one of the two decode functions. The function decode32 will store the pixel data
in 32 bit RGBA format, no matter what color format was used in the PNG file. The function decodeGeneric
will store the pixel data in the same color format as in the file.

Here is an example how you could use LodePNG with opengl:
---
uint loadPNG(char[] filename)
{
    uint textureID;

    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &textureID);
    glBindTexture(GL_TEXTURE_2D, textureID);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

    ubyte[] image;
    Info info = lodepng.decode32(image, cast(ubyte[])std.file.read(filename));
    if (info.error != 0)
        throw new Exception("Error loading png " ~
                            filename ~ ", errorcode: " ~
                            std.string.toString(info.error));

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, info.w, info.h, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                        image.ptr);
    return textureID;
}
---

Encoder:
The encoder currently produces a valid PNG, with a size much smaller than an uncompressed image.

With the current default settings, the encoder will expect a 32-bit RGBA buffer as input, but it'll automatically detect
if the image is fully opaque, and only include an alpha channel in the PNG if the image isn't opaque.


API:*/
module png;
import std.zlib;

/**This function converts the given compressed PNG data into uncompressed pixel
data. The original color type is preserved, which can make it more difficult
to interpret the pixel data since you have to take into account what color
type it is.

For normal usage of LodePNG, this function is *not* recommended, but in special
cases where you know beforehand what color type the PNG image has it can provide
an increase in efficiency. Use decode32 instead if you want any type of PNG
to be easily displayed! The decode32 function first calls decodeGeneric
and then does the job of converting all possible color types to RGBA.

Params:
_in = This is used to give the original file data.
 Provide an array filled with the contents of a PNG file. The contents of
 this buffer will be modified during the decoding process, so it cannot be
 reused. If you need the file contents a second time, make a copy before
 decoding the PNG or reload the file.

_out = This is used to return the pixel data.
 Provide an empty array. The vector will be resized and the uncompressed
 pixel data will be stored in the buffer. The image data will be of the same
 color type as the PNG image. Pixels are always stored row by row. The first
 byte of the data contains the first byte of the pixel of the top left of
 the image. 1-bit images or other formats with less than one byte per pixel
 will be stored bit per bit. In that case, scanlines don't necessarily start
 at the boundary of a  byte. If RGB or RGBA is used, color components per pixel
 are successive bits/bytes in the order R, G, B, A.
*/
Info decodeGeneric(inout ubyte[] _out, inout ubyte[] _in)
{
    Info info;
  //read the information from the header and store it in the Info
    void readPngHeader(ubyte* _in, size_t inlength)
    {
      if(inlength < 29) { info.error = 27; return; } //error: the data length is smaller than the length of the header

      if(_in[0] != 137 || _in[1] != 80 || _in[2] != 78 || _in[3] != 71 || _in[4] != 13 || _in[5] != 10 || _in[6] != 26 || _in[7] != 10) { info.error = 28; return; } //error: the first 8 bytes are not the correct PNG signature
      if(_in[12] != 73 || _in[13] != 72 || _in[14] != 68 || _in[15] != 82) { info.error = 29; return; } //error: it doesn't start with a IHDR chunk!

      //read the values given in the header
      info.w = 256 * 256 * 256 * _in[16] + 256 * 256 * _in[17] + 256 * _in[18] + _in[19];
      info.h = 256 * 256 * 256 * _in[20] + 256 * 256 * _in[21] + 256 * _in[22] + _in[23];
      info.bitDepth = _in[24];
      info.colorType = _in[25];
      info.compressionMethod = _in[26];
      info.filterMethod = _in[27];
      info.interlaceMethod = _in[28];
      //The 4 CRC bytes are ignored

      if(info.compressionMethod != 0) { info.error = 32; return; } //error: only compression method 0 is allowed in the specification
      if(info.filterMethod != 0)      { info.error = 33; return; } //error: only filter method 0 is allowed in the specification
      if(info.interlaceMethod > 1)    { info.error = 34; return; } //error: only interlace methods 0 and 1 exist in the specification

      //check if bit depth is valid for this color type, if not give error 37 "illegal bit depth for this color type given"
      uint bd = info.bitDepth; //longer variable name makes code below more easily readable
      switch(info.colorType)
      {
        case 0: if(bd != 1 && bd != 2 && bd != 4 && bd != 8 && bd != 16) { info.error = 37; return; } break;
        case 2: if(                                 bd != 8 && bd != 16) { info.error = 37; return; } break;
        case 3: if(bd != 1 && bd != 2 && bd != 4 && bd != 8            ) { info.error = 37; return; } break;
        case 4: if(                                 bd != 8 && bd != 16) { info.error = 37; return; } break;
        case 6: if(                                 bd != 8 && bd != 16) { info.error = 37; return; } break;
        default: { info.error = 31; return; } break; //error: invalid color type
      }
    }

    //filter a PNG image scanline by scanline. when the pixels are smaller than 1 byte, the filter works byte per byte (bytewidth = 1)
    //precon is the previous filtered scanline, recon the result, scanline the current one
    void unFilterScanline(ubyte[] recon, ubyte[] scanline, ubyte[] precon, bool top, size_t bytewidth, uint filterType, size_t length)
    {
      switch(filterType)
      {
        case 0:
          for(size_t i = 0; i < length; i++) recon[i] = scanline[i];
          break;
        case 1:
          if(top)
          {
            for(size_t i = 0; i < bytewidth; i++) recon[i] = scanline[i];
            for(size_t i = bytewidth; i < length; i++) recon[i] = scanline[i] + recon[i - bytewidth];
          }
          else
          {
            for(size_t i =         0; i < bytewidth; i++) recon[i] = scanline[i];
            for(size_t i = bytewidth; i < length   ; i++) recon[i] = scanline[i] + recon[i - bytewidth];
          }
          break;
        case 2:
          if(top) for(size_t i = 0; i < length; i++) recon[i] = scanline[i];
          else for(size_t i = 0; i < length; i++) recon[i] = scanline[i] + precon[i];
          break;
        case 3:
          if(top)
          {
            for(size_t i = 0; i < length; i++) recon[i] = scanline[i];
            for(size_t i = bytewidth; i < length; i++) recon[i] = scanline[i] + recon[i - bytewidth] >> 1;
          }
          else
          {
            for(size_t i = 0; i < bytewidth; i++) recon[i] = scanline[i] + precon[i] / 2;
            for(size_t i = bytewidth; i < length; i++) recon[i] = scanline[i] + ((recon[i - bytewidth] + precon[i]) >> 1);
          }
          break;
        case 4:
          if(top)
          {
            for(size_t i = 0; i < bytewidth; i++) recon[i] = scanline[i];
            for(size_t i = bytewidth; i < length; i++) recon[i] = cast(ubyte)(scanline[i] + paethPredictor(recon[i - bytewidth], 0, 0));
          }
          else
          {
            for(size_t i = 0; i < bytewidth; i++) recon[i] = cast(ubyte)(scanline[i] + paethPredictor(0, precon[i], 0));
            for(size_t i = bytewidth; i < length; i++) recon[i] = cast(ubyte)(scanline[i] + paethPredictor(recon[i - bytewidth], precon[i], precon[i - bytewidth]));
          }
          break;
      default: info.error = 36; return; //error: unexisting filter type given
      }
    }

    //filter and reposition the pixels into the output when the image is Adam7 interlaced. This function can only do it after the full image is already decoded. The out buffer must have the correct allocated memory size already.
    void adam7Pass(ubyte[] _out, ubyte[] _in, ubyte[] scanlinen, ubyte[] scanlineo, size_t w, size_t bytewidth, size_t passleft, size_t passtop, size_t spacex, size_t spacey, size_t passw, size_t passh, uint bpp)
    {
      for(uint s = 0; s < passh; s++)
      {
        size_t linelength = 1 + ((bpp * passw + 7) >> 3);
        size_t linestart = s * linelength; //position where we read the filterType: at the start of the scanline
        uint filterType = _in[linestart];

        unFilterScanline(scanlinen, _in[linestart + 1..$], scanlineo, (s < 1), bytewidth, filterType, (w * bpp + 7) >> 3);
        if(info.error) return;

        //put the filtered pixels in the output image
        if(bpp >= 8)
        {
          for(size_t i = 0; i < passw; i++)
          for(size_t b = 0; b < bytewidth; b++) //b = current byte of this pixel
          {
            _out[bytewidth * w * (passtop + spacey * s) + bytewidth * (passleft + spacex * i) + b] = scanlinen[bytewidth * i + b];
          }
        }
        else
        {
          for(size_t i = 0; i < passw; i++)
          {
            size_t outbitp = bpp * w * (passtop + spacey * s) + bpp * (passleft + spacex * i);
            for(size_t b = 0; b < bpp; b++) //b = current bit of this pixel
            {
              size_t obp = outbitp + b;
              size_t obitpos = 7 - (obp & 0x7); //where bitpos 0 refers to the LSB, bitpot 7 to the MSB of a byte
              size_t bp = i * bpp + b;
              size_t bitpos = 7 - (bp & 0x7); //where bitpos 0 refers to the LSB, bitpot 7 to the MSB of a byte
              uint _bit = (scanlinen[bp >> 3] >> bitpos) & 1;
              _out[obp >> 3] = cast(ubyte)((_out[obp >> 3] & ~(1 << obitpos)) | (_bit << obitpos));
            }
          }
        }

        //swap the two buffer pointers "scanline old" and "scanline new"
        ubyte[] temp = scanlinen;
        scanlinen = scanlineo;
        scanlineo = temp;
      }
    }

    void resetParameters()
    {
      info.error = 0; //initially no error happened yet
      info.paletteSize = info.backgroundColor = info.colorKey = 0; //initialize info variables that aren't necessarily set later on
    }

  //if(CHAR_BIT < 8) { info.error = 54; return; } // not relevant in D, type sizes are guaranteed
  if(_in.length == 0 && _out.length > 0) { info.error = 56; return info; } //the given input data is empty but the output data not
  if(_in.length == 0) { info.error = 48; return info; } //the given data is empty

  resetParameters(); //when decoding a new PNG image, make sure all parameters created after previous decoding are reset

  readPngHeader(_in.ptr, _in.length);
  if(info.error) return info;

  size_t pos = 33; //first byte of the first chunk after the header
  size_t dataPos = 0; //at this position in the in buffer the new data will be stored.
  size_t dataStart = 0;

  bool IEND = 0;
  while(!IEND) //loop through the chunks, ignoring unknown chunks and stopping at IEND chunk. IDAT data is put at the start of the in buffer
  {
    //get chunk length
    if(pos + 8 >= _in.length) { info.error = 30; return info; } //error: size of the in buffer too small to contain next chunk
    size_t chunkLength = 256 * 256 * 256 * _in[pos] + 256 * 256 * _in[pos + 1] + 256 * _in[pos + 2] + _in[pos + 3]; pos += 4;
    if(pos + chunkLength >= _in.length) { info.error = 35; return info; } //error: size of the in buffer too small to contain next chunk

    //IDAT chunk, containing compressed image data
    if(_in[pos + 0] == 'I' && _in[pos + 1] == 'D' && _in[pos + 2] == 'A' && _in[pos + 3] == 'T')
    {
      pos += 4;
      if(dataPos == 0) { dataStart = pos; dataPos = dataStart + chunkLength; pos += chunkLength; } //possible efficiency increase by not copying data from first chunk
      else for(size_t i = 0; i < chunkLength; i++) _in[dataPos++] = _in[pos++]; //multiple data chunks are added behind each other this way
    }
    //IEND chunk
    else if(_in[pos + 0] == 'I' && _in[pos + 1] == 'E' && _in[pos + 2] == 'N' && _in[pos + 3] == 'D')
    {
      IEND = 1;
    }
    //palette chunk (PLTE)
    else if(_in[pos + 0] == 'P' && _in[pos + 1] == 'L' && _in[pos + 2] == 'T' && _in[pos + 3] == 'E')
    {
      pos += 4; //go after the 4 letters
      info.paletteSize = chunkLength / 3;
      if(info.paletteSize > 256) { info.error = 38; return info; } //error: palette too big
      info.palette.length = 4 * info.paletteSize;
      for(size_t i = 0; i < info.paletteSize; i++)
      {
        info.palette[i * 4 + 0] = _in[pos++]; //R
        info.palette[i * 4 + 1] = _in[pos++]; //G
        info.palette[i * 4 + 2] = _in[pos++]; //B
        info.palette[i * 4 + 3] = 255; //alpha
      }
    }
    //palette transparency chunk (tRNS)
    else if(_in[pos + 0] == 't' && _in[pos + 1] == 'R' && _in[pos + 2] == 'N' && _in[pos + 3] == 'S')
    {
      pos += 4; //go after the 4 letters
      if(info.colorType == 3)
      {
        if(chunkLength > info.paletteSize) { info.error = 39; return info; } //error: more alpha values given than there are palette entries
        for(size_t i = 0; i < chunkLength; i++) info.palette[i * 4 + 3] = _in[pos++];
      }
      else if(info.colorType == 0)
      {
        if(chunkLength != 2) { info.error = 40; return info; } //error: this chunk must be 2 bytes for greyscale image
        info.colorKey = 1;
        info.keyR = 256 * _in[pos] + _in[pos + 1]; pos += 2;
      }
      else if(info.colorType == 2)
      {
        if(chunkLength != 6) { info.error = 41; return info; } //error: this chunk must be 6 bytes for RGB image
        info.colorKey = 1;
        info.keyR = 256 * _in[pos] + _in[pos + 1]; pos += 2;
        info.keyG = 256 * _in[pos] + _in[pos + 1]; pos += 2;
        info.keyB = 256 * _in[pos] + _in[pos + 1]; pos += 2;
      }
      else { info.error = 42; return info; } //error: tRNS chunk not allowed for other color models
    }
    //background color chunk (bKGD)
    else if(_in[pos + 0] == 'b' && _in[pos + 1] == 'K' && _in[pos + 2] == 'G' && _in[pos + 3] == 'D')
    {
      pos += 4; //go after the 4 letters
      if(info.colorType == 3)
      {
        if(chunkLength != 1) { info.error = 43; return info; } //error: this chunk must be 1 byte for indexed color image
        info.backgroundColor = 1;
        info.backgroundR = _in[pos++];
      }
      else if(info.colorType == 0 || info.colorType == 4)
      {
        if(chunkLength != 2) { info.error = 44; return info; } //error: this chunk must be 2 bytes for greyscale image
        info.backgroundColor = 1;
        info.backgroundR = 256 * _in[pos] + _in[pos + 1]; pos += 2;
      }
      else if(info.colorType == 2 || info.colorType == 6)
      {
        if(chunkLength != 6) { info.error = 45; return info; } //error: this chunk must be 6 bytes for greyscale image
        info.backgroundColor = 1;
        info.backgroundR = 256 * _in[pos] + _in[pos + 1]; pos += 2;
        info.backgroundG = 256 * _in[pos] + _in[pos + 1]; pos += 2;
        info.backgroundB = 256 * _in[pos] + _in[pos + 1]; pos += 2;
      }
    }
    else //it's not an implemented chunk type, so ignore it: skip over the data and the CRC
    {
      pos += 4; //go after the 4 letters
      pos += chunkLength; //skip uninterpreted data of unimplemented chunk
    }
    pos += 4; //skip CRC
  }

  switch(info.colorType)
  {
    case 0: info.bpp = info.bitDepth    ; info.colorChannels = 1; break; //grey
    case 2: info.bpp = info.bitDepth * 3; info.colorChannels = 3; break; //RGB
    case 3: info.bpp = info.bitDepth    ; info.colorChannels = 1; break; //palette
    case 4: info.bpp = info.bitDepth * 2; info.colorChannels = 2; break; //grey + alpha
    case 6: info.bpp = info.bitDepth * 4; info.colorChannels = 4; break; //RGBA
    default: break; //can never happen since colorType is always one of the above and error is already checked for that
  }

  size_t scanlength = ((info.w * (info.h * info.bpp + 7)) >> 3) + info.h * 8; //scanline buffer length is larger than final image size because up to h * 7 filter type codes can still be in it!

  ubyte[] scanlines = cast(ubyte[])std.zlib.uncompress(_in[dataStart..$], scanlength);
  scanlines.length = scanlength;

  if(info.error) return info;

  //filter and interlace
  size_t bytewidth = (info.bpp + 7) >> 3; //bytewidth is used for filtering
  size_t outlength = (info.h * info.w * info.bpp + 7) >> 3;

  _out.length = outlength; //time to fill the out buffer

  if(info.interlaceMethod == 0)
  {

    size_t linestart = 0; //start of current scanline
    size_t obp = 0; //out bit pointer, only used if bpp < 8

    uint filterType = scanlines[linestart];

    if(info.bpp >= 8) //byte per byte
    {
        filterType = scanlines[linestart];
        assert(info.bpp >= 8);
        unFilterScanline(_out,
                    scanlines[1..$],
                    _out, true, bytewidth, filterType,
                    (info.w * info.bpp + 7) >> 3);
        if(info.error) return info;
        linestart += 1 + ((info.w * info.bpp + 7) >> 3); //go to start of next scanline

        for(size_t s = 1; s < info.h; s++)
        {
            filterType = scanlines[linestart];
            assert(info.bpp >= 8);
            unFilterScanline(_out[linestart - s..$],
                    scanlines[linestart + 1..$],
                    _out[(s - 1) * info.w * bytewidth..$], false,
                    bytewidth, filterType,
                    (info.w * info.bpp + 7) >> 3);
            if(info.error) return info;
            linestart += 1 + ((info.w * info.bpp + 7) >> 3); //go to start of next scanline
        }
    }
    else //less than 8 bits per pixel, so fill it up bit per bit
    {
      ubyte[] templine = new ubyte[(info.w * info.bpp + 7) >> 3]; //only used if bpp < 8
      filterType = scanlines[linestart];
      unFilterScanline(templine, scanlines[1..$], _out, (0 < 1), bytewidth, filterType,  (info.w * info.bpp + 7) >> 3);
      if(info.error) return info;

      for(size_t s = 1; s < info.h; s++)
      {
        filterType = scanlines[linestart];
        unFilterScanline(templine, scanlines[linestart + 1..$], _out[(s - 1) * info.w * bytewidth..$], false, bytewidth, filterType,  (info.w * info.bpp + 7) >> 3);
        if(info.error) return info;

        for(size_t bp = 0; bp < info.w * info.bpp; bp++) //bp is here bit pointer in templine
        {
          size_t obitpos = 7 - (obp & 0x7);
          size_t bitpos = 7 - (bp & 0x7);
          uint _bit = (templine[bp >> 3] >> bitpos) & 1;
          _out[obp >> 3] = cast(ubyte)((_out[obp >> 3] & ~(1 << obitpos)) | (_bit << obitpos)); //set current bit
          obp++;
        }

        linestart += 1 + ((info.w * info.bpp + 7) >> 3); //go to start of next scanline
      }
    }
  }
  else //interlaceMethod is 1 (Adam7)
  {
    size_t[7] passw, passh, passstart;

    passw[0] = (info.w + 7) / 8, passh[0] = (info.h + 7) / 8;
    passw[1] = (info.w + 3) / 8, passh[1] = (info.h + 7) / 8;
    passw[2] = (info.w + 3) / 4, passh[2] = (info.h + 3) / 8;
    passw[3] = (info.w + 1) / 4, passh[3] = (info.h + 3) / 4;
    passw[4] = (info.w + 1) / 2, passh[4] = (info.h + 1) / 4;
    passw[5] = (info.w + 0) / 2, passh[5] = (info.h + 1) / 2;
    passw[6] = (info.w + 0) / 1, passh[6] = (info.h + 0) / 2;

    passstart[0] = 0;
    passstart[1] =        passh[0] * ((passw[0] * info.bpp + 7) / 8) + passh[0];
    passstart[2] = passstart[1] + passh[1] * ((passw[1] * info.bpp + 7) / 8) + passh[1];
    passstart[3] = passstart[2] + passh[2] * ((passw[2] * info.bpp + 7) / 8) + passh[2];
    passstart[4] = passstart[3] + passh[3] * ((passw[3] * info.bpp + 7) / 8) + passh[3];
    passstart[5] = passstart[4] + passh[4] * ((passw[4] * info.bpp + 7) / 8) + passh[4];
    passstart[6] = passstart[5] + passh[5] * ((passw[5] * info.bpp + 7) / 8) + passh[5];

    ubyte[] scanlineo = new ubyte[info.w * info.bpp]; //"old" scanline
    ubyte[] scanlinen = new ubyte[info.w * info.bpp]; //"new" scanline

    adam7Pass(_out, scanlines[passstart[0]..$], scanlinen, scanlineo, info.w, bytewidth, 0, 0, 8, 8, passw[0], passh[0], info.bpp);
    adam7Pass(_out, scanlines[passstart[1]..$], scanlinen, scanlineo, info.w, bytewidth, 4, 0, 8, 8, passw[1], passh[1], info.bpp);
    adam7Pass(_out, scanlines[passstart[2]..$], scanlinen, scanlineo, info.w, bytewidth, 0, 4, 4, 8, passw[2], passh[2], info.bpp);
    adam7Pass(_out, scanlines[passstart[3]..$], scanlinen, scanlineo, info.w, bytewidth, 2, 0, 4, 4, passw[3], passh[3], info.bpp);
    adam7Pass(_out, scanlines[passstart[4]..$], scanlinen, scanlineo, info.w, bytewidth, 0, 2, 2, 4, passw[4], passh[4], info.bpp);
    adam7Pass(_out, scanlines[passstart[5]..$], scanlinen, scanlineo, info.w, bytewidth, 1, 0, 2, 2, passw[5], passh[5], info.bpp);
    adam7Pass(_out, scanlines[passstart[6]..$], scanlinen, scanlineo, info.w, bytewidth, 0, 1, 1, 2, passw[6], passh[6], info.bpp);
  }
  return info;
}

/** This function converts the given compressed PNG data into uncompressed 32-bit RGBA pixel data, row by row.
    The pixel data is always in 32-bit RGBA, no matter what color type was used in the PNG image. This makes it rather easy to
    create textures or plot pixels from any kind of PNG file.

    Params:
_in =  This is used to give the original file data.
 Provide a vector filled with the contents of a PNG file. The contents of
 this buffer will be modified during the decoding process, so it cannot be
 reused. If you need the file contents a second time, make a copy before
 decoding the PNG or reload the file.

_out = This is used to return the pixel data.
 Provide an empty vector. The vector will be resized and the uncompressed
 pixel data will be stored in the buffer. The pixel data will be of the RGBA
 format with 32 bits (4 bytes) per pixel. Pixels are stored row by row. The
 first four bytes of the data contain the four bytes of the pixel of the top
 left of the image. The next four bytes contain the second pixel, and so on.
 The first byte of a pixel is the R component, the second the G component, the
 third the B component and the fourth the A component. The size of the out
 buffer will be the number of pixels in the image times four bytes.
*/
Info decode32(inout ubyte[] _out, ubyte[] _in)
{
  Info info;

  if(_in.length == 0 && _out.length > 0) { info.error = 56; return info; } //the given input data is empty but the output data not
  ubyte[] buffer;
  info = decodeGeneric(buffer, _in); //before we can convert to RGBA, decode the PNG in the format it is
  if(info.error) return info;

  size_t w = info.w; //width of the image
  size_t h = info.h; //height of the image

  _out.length = w * h * 4;

  switch(info.colorType)
  {
    case 0: //greyscale color
    for(size_t i = 0; i < w * h; i++)
    {
      _out[4 * i + 3] = 255;
      if(info.bitDepth == 8)
      {
        _out[4 * i + 0] = _out[4 * i + 1] = _out[4 * i + 2] = buffer[i];
        if(info.colorKey == 1)
        if(buffer[i] == info.keyR) _out[4 * i + 3] = 0;
      }
      else if(info.bitDepth == 16)
      {
        _out[4 * i + 0] = _out[4 * i + 1] = _out[4 * i + 2] = buffer[2 * i];
        if(info.colorKey == 1)
        if(256U * buffer[i] + buffer[i + 1] == info.keyR) _out[4 * i + 3] = 0;
      }
      else
      {
        size_t bp = info.bitDepth * i;
        uint value = 0;
        uint pot = 1 << info.bitDepth; //power of two
        for(size_t j = 0; j < info.bitDepth; j++)
        {
          pot /= 2;
          uint _bit = readBitFromStreamReversed(bp, buffer.ptr);
          value += pot * _bit;
        }
        if(info.colorKey == 1)
        if(value && ((1U << info.bitDepth) - 1U) == info.keyR && ((1U << info.bitDepth) - 1U)) _out[4 * i + 3] = 0;
        value = (value * 255) / ((1 << info.bitDepth) - 1); //scale value from 0 to 255
        _out[4 * i + 0] = _out[4 * i + 1] = _out[4 * i + 2] = cast(ubyte)(value);
      }
    }
    break;
    case 2: //RGB color
    for(size_t i = 0; i < w * h; i++)
    {
      _out[4 * i + 3] = 255;
      if(info.bitDepth == 8)
      {
        for(size_t c = 0; c < 3; c++) _out[4 * i + c] = buffer[3 * i + c];
        if(info.colorKey == 1)
        if(buffer[3 * i + 0] == info.keyR && buffer[3 * i + 1] == info.keyG && buffer[3 * i + 2] == info.keyB) _out[4 * i + 3] = 0;
      }
      else if(info.bitDepth == 16)
      {
        for(size_t c = 0; c < 3; c++) _out[4 * i + c] = buffer[6 * i + 2 * c];
        if(info.colorKey == 1)
        if(256U * buffer[6 * i + 0] + buffer[6 * i + 1] == info.keyR
        && 256U * buffer[6 * i + 2] + buffer[6 * i + 3] == info.keyG
        && 256U * buffer[6 * i + 4] + buffer[6 * i + 5] == info.keyB) _out[4 * i + 3] = 0;
      }
    }
    break;
    case 3: //indexed color (palette)
    for(size_t i = 0; i < w * h; i++)
    {
      _out[4 * i + 3] = 255;
      if(info.bitDepth == 8)
      {
        uint value = buffer[i];
        if(value >= info.paletteSize) { info.error = 46; return info; }
        for(size_t c = 0; c < 4; c++) _out[4 * i + c] = cast(ubyte)(info.palette[c + 4 * value]); //get rgb colors from the palette, which is stored in info
      }
      else if(info.bitDepth < 8)
      {
        size_t bp = info.bitDepth * i;
        uint value = 0;
        uint pot = 1 << info.bitDepth; //power of two
        for(size_t j = 0; j < info.bitDepth; j++)
        {
          pot /= 2;
          uint _bit = readBitFromStreamReversed(bp, buffer.ptr);
          value += pot * _bit;
        }
        if(value >= info.paletteSize) { info.error = 47; return info; }
        for(size_t c = 0; c < 4; c++) _out[4 * i + c] = info.palette[c + 4 * value]; //get rgb colors from the palette, which is stored in info
      }
    }
    break;
    case 4: //greyscale with alpha
    for(size_t i = 0; i < w * h; i++)
    {
      _out[4 * i + 3] = 255;
      if(info.bitDepth == 8)
      {
        _out[4 * i + 0] = _out[4 * i + 1] = _out[4 * i + 2] = buffer[i * 2 + 0];
        _out[4 * i + 3] = buffer[i * 2 + 1];
      }
      else if(info.bitDepth == 16)
      {
        _out[4 * i + 0] = _out[4 * i + 1] = _out[4 * i + 2] = buffer[4 * i]; //most significant byte
        _out[4 * i + 3] = buffer[4 * i + 2]; //most significant byte
      }
    }
    break;
    case 6: //RGB with alpha
    for(size_t i = 0; i < w * h; i++)
    {
      _out[4 * i + 3] = 255;
      if   (info.bitDepth ==  8) for(size_t c = 0; c < 4; c++) _out[4 * i + c] = buffer[4 * i + c];
      else if(info.bitDepth == 16) for(size_t c = 0; c < 4; c++) _out[4 * i + c] = buffer[8 * i + 2 * c];
    }
    break;
    default: break;
  }
  return info;
}

/**Information about the PNG image is returned in the form of an Info struct.

After using decodeGeneric or decode32, please always check the error value
to see if everything went ok.
*/
struct Info
{
    /** Below is a table that gives the meaning of the different values.

    In the port to D, some of these errors have become obsolete as zlib has replaced a few routines and will spit out
    exceptions instead. I have not yet removed these.

<li> 0: no error, everything went ok
<li> 1: there is no PNG loaded yet, so it makes no sense trying to read any information values
<li> 10: while huffman decoding: end of input memory reached without endcode
<li> 11: while huffman decoding: error in code tree made it jump outside of tree
<li> 13: problem while processing dynamic deflate block
<li> 14: problem while processing dynamic deflate block
<li> 15: problem while processing dynamic deflate block
<li> 16: unexisting code while processing dynamic deflate block
<li> 17: while inflating: end of out buffer memory reached
<li> 18: while inflating: invalid distance code
<li> 19: while inflating: end of out buffer memory reached
<li> 20: invalid deflate block BTYPE encountered
<li> 21: NLEN is not ones complement of LEN in a deflate block
<li> 22: while inflating: end of out buffer memory reached.
   This can happen if the inflated deflate data is longer than the amount of bytes required to fill up
   all the pixels of the image, given the color depth and image dimensions. Something that doesn't
   happen in a normal, well encoded, PNG image.
<li> 23: while inflating: end of in buffer memory reached
<li> 24: invalid FCHECK in zlib header
<li> 25: invalid compression method in zlib header
<li> 26: FDICT encountered in zlib header while it's not used for PNG
<li> 27: PNG file is smaller than a PNG header
<li> 28: incorrect PNG signature (the first 8 bytes of the PNG file)
   Maybe it's not a PNG, or a PNG file that got corrupted so that the header indicates the corruption.
<li> 29: first chunk is not the header chunk
<li> 30: chunk length too large, chunk broken off at end of file
<li> 31: illegal PNG color type
<li> 32: illegal PNG compression method
<li> 33: illegal PNG filter method
<li> 34: illegal PNG interlace method
<li> 35: chunk length of a chunk is too large or the chunk too small
<li> 36: illegal PNG filter type encountered
<li> 37: illegal bit depth for this color type given
<li> 38: the palette is too big (more than 256 colors)
<li> 39: more palette alpha values given in tRNS, than there are colors in the palette
<li> 40: tRNS chunk has wrong size for greyscale image
<li> 41: tRNS chunk has wrong size for RGB image
<li> 42: tRNS chunk appeared while it was not allowed for this color type
<li> 43: bKGD chunk has wrong size for palette image
<li> 44: bKGD chunk has wrong size for greyscale image
<li> 45: bKGD chunk has wrong size for RGB image
<li> 46: value encountered in indexed image is larger than the palette size (bitdepth == 8)
<li> 47: value encountered in indexed image is larger than the palette size (bitdepth < 8)
<li> 48: the input data is empty. Maybe a PNG file you tried to load doesn't exist or is in the wrong path.
<li> 49: jumped past memory while generating dynamic huffman tree
<li> 50: jumped past memory while generating dynamic huffman tree
<li> 51: jumped past memory while inflating huffman block
<li> 52: jumped past memory while inflating
<li> 53: size of zlib data too small to contain header
<li> 54: CHAR_BITS is smaller than 8 on this platform, an 8-bit datatype is needed to store the file and pixel data
<li>55: jumped past tree while generating huffman tree, this could be when the code
       lengths are not of an optimal_ tree, which causes there to be more nodes than
       the arrays used can support (e.g. if there are 19 codes a tree has 18 nodes,
       but if all 19 code lengths are 7, there will be much more nodes), this may be a
       BUG that has to be fixed though I haven't actually encountered PNGs causing
       this problem outside the experimentation environment
<li> 56: As of LodePNG version 20061209, the input and output parameter of the decode functions got swapped!
       Try to swap the input and output parameter in the decode function call. (first out, then in)
       If they are already correct and you get this error, then see description of error 48.
    */
    int error = 1; //the error value of the decode attempt

    // dimensions of the image
    size_t w; /// width of the image in pixels
    size_t h; /// height of the image in pixels

    //pixel color info
    uint bitDepth;      /// bits per sample
    uint bpp;           /// bits per pixel
    uint colorChannels; /// amount of channels

    //palette info
    uint paletteSize; /// size of the palette (palette.size() / 4)
    ubyte[] palette;  /// palette in RGBARGBA... order

    //transparent color key info
    bool  colorKey; /// is a transparent color key given?
    ubyte keyR;     /// red/greyscale component of color key
    ubyte keyG;     /// green component of color key
    ubyte keyB;     /// blue component of color key

    //PNG specific information (general)
    uint colorType;         /// color type of the original PNG file
    uint compressionMethod; /// compression method of the original file
    uint filterMethod;      /// filter method of the original file
    uint interlaceMethod;   /// interlace method of the original file

    //PNG specific information (bKGD chunk)
    bool  backgroundColor; /// is a suggested background color given?
    ubyte backgroundR;     /// red component of sugg. background color
    ubyte backgroundG;     /// green component of sugg. background color
    ubyte backgroundB;     /// blue component of sugg. background color
}


//////////////////////////////////////////////////////////////////////////////
// Private Functions And Data                                               //
//////////////////////////////////////////////////////////////////////////////
private
{
    //Paeth predicter, used by one of the PNG filter types
    int paethPredictor(int a, int b, int c)
    {
      int p = a + b - c;
      int pa = p > a ? p - a : a - p;
      int pb = p > b ? p - b : b - p;
      int pc = p > c ? p - c : c - p;

      if(pa <= pb && pa <= pc) return a;
      else if(pb <= pc) return b;
      return c;
    }

    //////////////////////////////////////////////////////////////////////////////
    // Reading and writing single bits to byte stream                           //
    //////////////////////////////////////////////////////////////////////////////

    void addBitToStream(inout size_t bitpointer, inout ubyte[] bitstream, uint _bit)
    {
      if(bitpointer % 8 == 0) bitstream ~= 0; //add a new byte at the end
      uint shiftedbit = (_bit << (bitpointer % 8)); //earlier bit of huffman code is in a lesser significant bit of an earlier byte
      bitstream[$-1] |= shiftedbit;
      bitpointer++;
    }

    void addBitsToStream(inout size_t bitpointer, inout ubyte[] bitstream, uint value, size_t nbits)
    {
      for(size_t j = 0; j < nbits; j++)
      {
        uint _bit = (value >> j) & 1;
        addBitToStream(bitpointer, bitstream, _bit);
      }
    }

    void addBitsToStreamReversed(inout size_t bitpointer, inout ubyte[] bitstream, uint value, size_t nbits)
    {
      for(size_t j = 0; j < nbits; j++)
      {
        uint _bit = (value >> (nbits - 1 - j)) & 1;
        addBitToStream(bitpointer, bitstream, _bit);
      }
    }

    uint readBitFromStream(inout size_t bitpointer, ubyte* bitstream)
    {
      return (bitstream[bitpointer >> 3] >> (bitpointer++ & 0x7)) & 1;
    }

    uint readBitFromStreamReversed(inout size_t bitpointer, ubyte* bitstream)
    {
      return (bitstream[bitpointer >> 3] >> (7 - bitpointer++ & 0x7)) & 1;
    }

    uint readBitsFromStream(inout size_t bitpointer, ubyte* bitstream, int nbits)
    {
      uint pot = 1; //power of two
      uint result = 0;
      for(int i = 0; i < nbits; i++)
      {
        uint _bit = readBitFromStream(bitpointer, bitstream);
        result += _bit * pot;
        pot *= 2;
      }
      return result;
    }
}
/*
PNG ENCODER
*/
void concatUint(inout ubyte[] bytestream, uint num)
{
    bytestream.length = bytestream.length + 4;
    bytestream[$-4] = num >> 24;
    bytestream[$-3] = num >> 16;
    bytestream[$-2] = num >> 8;
    bytestream[$-1] = num;
}

void addChunk(inout ubyte[] _out, string chunkName, ubyte[] data, size_t len)
{
  if(chunkName.length != 4) return;
  assert(chunkName.length == 4);
  //1: length
  _out.concatUint(len);

  //2: chunk name (4 letters)
  _out.length = _out.length + 4;
  _out[$-4] = chunkName[0];
  _out[$-3] = chunkName[1];
  _out[$-2] = chunkName[2];
  _out[$-1] = chunkName[3];

  //3: the data
  _out.length = _out.length + len;
  _out[$ -len..$] = data;

  //4: CRC (of the chunkname characters and the data)
  uint CRC = std.zlib.crc32(0, _out[$ - len - 4 .. $]);
  _out.concatUint(CRC);
}

void writeSignature(inout ubyte[] _out)
{
  //8 bytes PNG signature
  _out ~= [137, 80, 78, 71, 13, 10, 26, 10];
}

void writeChunk_IHDR(inout ubyte[] _out, uint w, uint h, int bitDepth, int colorType)
{
  ubyte[] header = new ubyte[13];
  header.length = 0;

  //header
  header.concatUint(w);
  header.concatUint(h);
  header ~= bitDepth; //bit depth
  header ~= colorType; //color type
  header ~= 0; //compression method
  header ~= 0; //filter method
  header ~= 0; //interlace method

  addChunk(_out, "IHDR", header, header.length);
  delete header;
}

void writeChunk_tEXt(inout ubyte[] _out, string keyword, char[] textstring)
{
  //add text chunk
  char[] text = keyword.dup;
  text ~= 0;
  text = text ~ textstring;
  addChunk(_out, "tEXt", cast(ubyte[])text, text.length);
}

void writeChunk_IDAT(inout ubyte[] _out, ubyte[] data)
{
      _out.length = _out.length + 8;
      uint dataStart = _out.length;
      _out ~= cast(ubyte[])std.zlib.compress(data, settings.compressionLevel);
      _out[dataStart - 4] = 'I';
      _out[dataStart - 3] = 'D';
      _out[dataStart - 2] = 'A';
      _out[dataStart - 1] = 'T';
      _out[dataStart - 8] = (_out.length - dataStart) >> 24;
      _out[dataStart - 7] = (_out.length - dataStart) >> 16;
      _out[dataStart - 6] = (_out.length - dataStart) >> 8;
      _out[dataStart - 5] = (_out.length - dataStart);
      uint CRC = std.zlib.crc32(0, _out[(_out.length - dataStart) - 1 .. $]);
      _out.concatUint(CRC);
}

void writeChunk_IEND(inout ubyte[] _out)
{
  addChunk(_out, "IEND", [], 0);
}

void dontFilter(inout ubyte[] _out, ubyte[] image, size_t w, size_t h)
{
  //currently bytewidth is only correctly calculated for the case of 8 bits per color channel
  size_t bytewidth;
  switch(settings.colorType)
  {
    case 0: bytewidth = 1; break;
    case 2: bytewidth = 3; break;
    case 3: bytewidth = 1; break;
    case 4: bytewidth = 2; break;
    case 6: bytewidth = 4; break;
    default: return; //this may not happen
  };

  _out.length = image.length + h;
  //generate the literal data _out of given image vector. filterType has to be added per scanline.
  for(size_t y = 0; y < h; y++)
  {
    size_t begin = y * (w * bytewidth + 1);
    _out[begin] = 0; //filterType 0 for this scanline
    for(size_t x = 0; x < w * bytewidth; x++)
      _out[begin + 1 + x] = image[y * w * bytewidth + x];
  }
}
/+ For adaptive filtering. Currently only type 4 is implemented (inlined, see filter())
void filterScanline(inout ubyte[] _out, ubyte[] scanline, ubyte[] prevline, bool top, size_t length, size_t bytewidth, int filterType)
{
  switch(filterType)
  {
    case 0:
      _out[] = scanline;
    case 1:
      if(top)
      {
        for(size_t i = 0; i < bytewidth; i++) _out[i] = scanline[i];
        for(size_t i = bytewidth; i < length; i++) _out[i] = scanline[i] - scanline[i - bytewidth];
      }
      else
      {
        for(size_t i =         0; i < bytewidth; i++) _out[i] = scanline[i];
        for(size_t i = bytewidth; i < length   ; i++) _out[i] = scanline[i] - scanline[i - bytewidth];
      }
      break;
    case 2:
      if(top) for(size_t i = 0; i < length; i++) _out[i] = scanline[i];
      else for(size_t i = 0; i < length; i++) _out[i] = scanline[i] - prevline[i];
      break;
    case 3:
      if(top)
      {
        for(size_t i = 0; i < length; i++) _out[i] = scanline[i];
        for(size_t i = bytewidth; i < length; i++) _out[i] = scanline[i] - scanline[i - bytewidth] / 2;
      }
      else
      {
        for(size_t i = 0; i < bytewidth; i++) _out[i] = scanline[i] - prevline[i] / 2;
        for(size_t i = bytewidth; i < length; i++) _out[i] = scanline[i] - ((scanline[i - bytewidth] + prevline[i]) / 2);
      }
      break;
    case 4:
      if(top)
      {
        for(size_t i = 0; i < bytewidth; i++) _out[i] = scanline[i];
        for(size_t i = bytewidth; i < length; i++) _out[i] = cast(ubyte)(scanline[i] - paethPredictor(scanline[i - bytewidth], 0, 0));
      }
      else
      {
        for(size_t i = 0; i < bytewidth; i++) _out[i] = cast(ubyte)(scanline[i] - paethPredictor(0, prevline[i], 0));
        for(size_t i = bytewidth; i < length; i++)
          _out[i] = cast(ubyte)(scanline[i] -
                               paethPredictor( scanline[i - bytewidth], prevline[i], prevline[i - bytewidth] )
                               );
      }
      break;
  default: return; //unexisting filter type given
  }
}+/

void filter(inout ubyte[] _out, inout ubyte[] image, size_t w, size_t h)
{
   _out.length = image.length + h; //image size plus an extra byte per scanline

  //currently bytewidth is only correctly calculated for the case of 8 bits per color channel
  uint bytewidth;

  switch(settings.colorType)
  {
    case 0: bytewidth = 1; break;
    case 2: bytewidth = 3; break;
    case 3: bytewidth = 1; break;
    case 4: bytewidth = 2; break;
    case 6: bytewidth = 4; break;
    default: return; //this may not happen
  }
  uint scanlength = w * bytewidth;
  assert(settings.colorType ==2 || settings.colorType == 6);

  ubyte[] result, current, previous;

  _out[0] = 4;
  result = _out[1..scanlength + 1];
  current = image[0..scanlength];

  for(size_t i = 0; i < bytewidth; i++)
    result[i] = current[i];
  for(size_t i = bytewidth; i < scanlength; i++)
    result[i] = cast(ubyte)(current[i] - paethPredictor(current[i - bytewidth], 0, 0));

  for (int y = 1; y < h; y++)
  {
      _out[y * (scanlength + 1)] = 4;
      result = _out[ y * (scanlength + 1) + 1 .. y * (scanlength + 1) + scanlength + 1];
      current = image[y * scanlength..y * scanlength + scanlength];
      previous = image[(y -1) * scanlength..(y - 1) * scanlength + scanlength];
      for(size_t i = 0; i < bytewidth; i++)
        result[i] = cast(ubyte)( current[i] - paethPredictor(0, previous[i], 0));

      for(size_t i = bytewidth; i < scanlength; i++)
        result[i] = cast(ubyte)( current[i] - paethPredictor( current[i - bytewidth], previous[i], previous[i - bytewidth] ) );
  }
}

void convertColorType(inout ubyte[] _out, int colorType_out, int bitDepth_out, ubyte[] _in, int colorType_in, int bitDepth_in)
{
  if(colorType_out == colorType_in && bitDepth_out == bitDepth_in)
  {
    _out = _in;
    return;
  }

  //the only other supported case now, is converting from RGBA to RGB, with 8 bits per channel
  if(colorType_in == 6 && colorType_out == 2)
  {
    size_t numpixels = _in.length / 4;
    _out.length = numpixels * 3;
    for(size_t i = 0; i < numpixels; i++)
    {
      int index_out = i * 3, index_in = i * 4;
      _out[index_out + 0] = _in[index_in + 0];
      _out[index_out + 1] = _in[index_in + 1];
      _out[index_out + 2] = _in[index_in + 2];
    }
  }
}

bool isFullyOpaque(ubyte[] image, uint colorType_in, uint bitDepth_in)
{
  if(bitDepth_in != 8) return false; //currently bitDepths other than 8 aren't supported

  if(colorType_in == 6)
  {
    size_t numpixels = image.length / 4;
    for(size_t i = 0; i < numpixels; i++)
    {
      if(image[i * 4 + 3] != 255) return false;
    }
  }
  else if(colorType_in == 2)
  {
    size_t numpixels = image.length / 2;
    for(size_t i = 0; i < numpixels; i++)
    {
      if(image[i * 2 + 1] != 255) return false;
    }
  }

  return true; //no single pixel with alpha channel other than 255 found
}

/// Encode a RGBA image in the PNG format.
/**
Currently encode accepts only RGBA images where each color channel has a bit depth of 8. By default, if the source image has
an alpha channel but it is fully opaque, the alpha channel is stripped and the result image is in the RGB format. You can
optionally provide a settings struct for several options.
*/
void encode(inout ubyte[] _out, ubyte[] image, size_t w, size_t h, Settings config = Settings.init)
{
  settings = config; //set the working settings to the user settings (the working settings might be modified while working)

  if(settings.autoLeaveOutAlphaChannel && isFullyOpaque(image, 6, 8))
  {
    //go to a color type without alpha channel
    if(settings.colorType == 6) settings.colorType = 2;
    else if(settings.colorType == 4) settings.colorType = 0;
  }

  ubyte[] converted;
  convertColorType(converted, settings.colorType, 8, image, 6, 8);

  ubyte[] data;
  assert(converted.length);
  filter(data, converted, w, h);
  writeSignature(_out);
  writeChunk_IHDR(_out, w, h, 8, settings.colorType);
  writeChunk_IDAT(_out, data);

  foreach(keyword, text; settings.textMap)
    writeChunk_tEXt(_out, keyword, text);
  writeChunk_IEND(_out);

}

private Settings settings;

/// Controls how a png image is encoded.
struct Settings
{
    static Settings opCall(uint compressionLevel = 6, uint colorType = 6, uint bitDepth = 8, bool autoLeaveOutAlphaChannel = true)
    {
        Settings result;
        result.compressionLevel = compressionLevel;
        result.colorType = colorType;
        result.bitDepth = bitDepth;
        result.autoLeaveOutAlphaChannel = autoLeaveOutAlphaChannel;
        return result;
    }
    invariant()
    {
        assert(compressionLevel >=0 && compressionLevel <= 9);
    }

    /// Zlib compression level (0 - 9). Higher compression levels may take much longer, while not reducing size that much. Defaults to 6.
    uint compressionLevel = 6;
    uint colorType = 6; //color type of resulting PNG
    uint bitDepth = 8; //bits per pixel of resulting PNG
    /// If set to true, a fully opaque alpha channel is stripped off. Default is true.
    bool autoLeaveOutAlphaChannel = true; //automatically use color type without alpha instead of given one, if given image is opaque
    /// Associative array of keyword - text pairs. Will write tEXt chunks.
    /**
    BUG: does not play nice with unicode, be certain that it contains only ascii characters!
    */
    char[][char[]] textMap;
}