//----------------------------------------------------------------------------
// RNA to PNG converter
//   Usage:
//     r2p in.png out.rna [lim]
//----------------------------------------------------------------------------

import std.stream;
import std.cstream;
import std.conv;
import std.string;
import png;

struct RGBA
{
	ubyte r;
	ubyte g;
	ubyte b;
	ubyte a;
};

alias RGBA[600][] Bitmap;

class Canvas
{
	int bR=0, bG=0, bB=0, bA=0, bN=0, bAN=0;
	int pos_x  = 0;
	int pos_y  = 0;
	int mark_x = 0;
	int mark_y = 0;
	int dir_x  = 1;
	int dir_y  = 0;
	Bitmap[] bmp;

	this() { addBitmap(); }
	void addBitmap() { if(bmp.length<10) bmp = new RGBA[600][600] ~ bmp; }
	void addColor(int r, int g, int b) {bR+=r;bG+=g;bB+=b;bN++;}
	void addAlpha(int a) {bA+=a; bAN++;}
	void clearColor(){bR=bG=bB=bA=bN=bAN=0;}
	void mark() { mark_x=pos_x; mark_y=pos_y; }
	void move() { pos_x=(pos_x+600+dir_x)%600; pos_y=(pos_y+600+dir_y)%600; }
	void turnLeft() { int nx=dir_y,ny=-dir_x; dir_x=nx; dir_y=ny; }
	void turnRight() { int nx=-dir_y,ny=dir_x; dir_x=nx; dir_y=ny; }
	void line()
	{
		RGBA p = currentPixel();
		int dx = mark_x - pos_x;
		int dy = mark_y - pos_y;
		int d = max(abs(dx),abs(dy));
		int c = (dx*dy<=0 ? 1 : 0);
		int x = pos_x*d + (d-c)/2;
		int y = pos_y*d + (d-c)/2;
		for(int i=0; i<d; ++i) {
			bmp[0][y/d][x/d] = p;
			x+=dx;
			y+=dy;
		}
		bmp[0][mark_y][mark_x] = p;
	}
	void fill()
	{
		RGBA nw = currentPixel();
		RGBA od = bmp[0][pos_y][pos_x];
		if( nw != od )
			fill_impl(pos_x, pos_y, nw, od);
	}
	void compose()
	{
		if( bmp.length >= 2 )
		{
			for(int y=0; y<600; ++y)
			for(int x=0; x<600; ++x)
			{
				RGBA z = bmp[0][y][x];
				RGBA o = bmp[1][y][x];
				bmp[1][y][x] = RGBA(z.r+o.r*(255-z.a)/255,
					z.g+o.g*(255-z.a)/255,
					z.b+o.b*(255-z.a)/255,
					z.a+o.a*(255-z.a)/255 );
			}
			bmp = bmp[1..$];
		}
	}
	void clip()
	{
		if( bmp.length >= 2 )
		{
			for(int y=0; y<600; ++y)
			for(int x=0; x<600; ++x)
			{
				int a = bmp[0][y][x].a;
				RGBA o = bmp[1][y][x];
				bmp[1][y][x] = RGBA(o.r*a/255,o.g*a/255,o.b*a/255,o.a*a/255 );
			}
			bmp = bmp[1..$];
		}
	}

	ubyte[] generatePNG()
	{
		ubyte[] png_data;
		ubyte[] raw;
		for(int y=0; y<600; ++y)
		for(int x=0; x<600; ++x) {
			raw ~= bmp[0][y][x].r;
			raw ~= bmp[0][y][x].g;
			raw ~= bmp[0][y][x].b;
			raw ~= 255;
		}
		png.encode(png_data, raw, 600, 600);
		return png_data;
	}



private:
	static int min(int a, int b) { return a<b ? a : b; }
	static int max(int a, int b) { return a>b ? a : b; }
	static int abs(int a) { return a>0 ? a : -a; }
	RGBA currentPixel()
	{
		ubyte a = bAN?bA/bAN:255;
		return RGBA((bN?bR/bN:0)*a/255, (bN?bG/bN:0)*a/255, (bN?bB/bN:0)*a/255, a);
	}
	void fill_impl(int x, int y, RGBA nw, RGBA od)
	{
		int[] sy=[y];
		int[] sx=[x];

		while(sx.length)
		{
			y = sy[$-1]; sy.length=sy.length-1;
			x = sx[$-1]; sx.length=sx.length-1;
			if( bmp[0][y][x] == od )
			{
				bmp[0][y][x] = nw;
				if(x>0&&bmp[0][y][x-1]==od)   {sx~=x-1; sy~=y;}
				if(x<599&&bmp[0][y][x+1]==od) {sx~=x+1; sy~=y;}
				if(y>0&&bmp[0][y-1][x]==od)   {sx~=x; sy~=y-1;}
				if(y<599&&bmp[0][y+1][x]==od) {sx~=x; sy~=y+1;}
			}
		}
	}
}

void main( string[] argv )
{
	if( argv.length <= 2 )
	{
		derr.writefln();
		derr.writefln("Usage:");
		derr.writefln("  r2p in.rna out.png [limit]" );
	}
	else
		with(new Canvas)
		{
			int lim = int.max;
			int c = 1;
			if(argv.length>=4) lim = toInt(argv[3]);
string prev="XXXXXXX";
			foreach(ulong i,string inst; new File(argv[1], FileMode.In))
			{
				void savePNG()
				  { std.file.write( format("%07d%s", i, argv[2]), generatePNG() ); }

				if(i%1000==0) derr.writefln("Instruction ", i, " processed.");
				if(i>=lim) break;
				switch(inst)
				{
				case "PIPIIIC": addColor(  0,  0,  0); break;
				case "PIPIIIP": addColor(255,  0,  0); break;
				case "PIPIICC": addColor(  0,255,  0); break;
				case "PIPIICF": addColor(255,255,  0); break;
				case "PIPIICP": addColor(  0,  0,255); break;
				case "PIPIIFC": addColor(255,  0,255); break;
				case "PIPIIFF": addColor(  0,255,255); break;
				case "PIPIIPC": addColor(255,255,255); break;
				case "PIPIIPF": addAlpha(  0); break;
				case "PIPIIPP": addAlpha(255); break;
				case "PIIPICP": clearColor(); break;
				case "PIIIIIP": move(); break;
				case "PCCCCCP": turnLeft(); break;
				case "PFFFFFP": turnRight(); break;
				case "PCCIFFP": mark(); break;
				case "PFFICCP": line(); break;
				case "PIIPIIP": fill(); break;
				case "PCCPFFP": addBitmap(); break;
				case "PFFPCCP": /* savePNG(); dout.writefln("snapshot");*/ compose(); break;
				case "PFFICCF": clip(); break;
				default: break;
				}

				//if(prev[0]=='C' && inst=="PIIPICP")
				//	dout.writefln(prev);
prev = inst.dup;
			}

			std.file.write( argv[2], generatePNG() );
		}
}
